#include <RTClib.h>
#include <Adafruit_GPS.h>
#include "Wire.h"
#include <Time.h>  
#include <SoftwareSerial.h>
#define DS3231_I2C_ADDRESS 0x68
#define TCAADDR 0x70

#define REFERENCE_RTC 0
#define GPS_RTC 1
#define ATOMIC_RTC 2

// Set to false to display time in 12 hour format, or true to use 24 hour:
#define TIME_24_HOUR      true

// Offset the hours from UTC (universal time) to your local time by changing
// this value.  The GPS time will be in UTC so lookup the offset for your
// local time from a site like:
//   https://en.wikipedia.org/wiki/List_of_UTC_time_offsets
// This value, -7, will set the time to UTC-7 or Pacific Standard Time during
// daylight savings time.
#define HOUR_OFFSET       -7

#define I     254 //Instruction prefix value
#define CLR   1 //LCD clear-screen instruction
#define LINE1 128 //Address of first character of 1st line
#define LINE2 192 //Address of first character of 2nd line

// Attach the serial display's RX line to digital pin 11
SoftwareSerial lcdSerial(7,11,1); // pin 11 = TX, pin 10 = RX (unused), inverted
SoftwareSerial gpsSerial(9,10); // pin 1 = TX, pin 10 = RX (unused), inverted

Adafruit_GPS gps(&gpsSerial);

// confugure interrupts to the pps clock signals
volatile unsigned long count = 0;
volatile bool gps_flag = false;
volatile bool atomic_flag = false;
volatile unsigned long rtc_start_ms;
volatile unsigned long pps_start_ms;

int setTime = false;

void setup()
{
  delay(500); // wait for display/gps to boot up
  // Setup the GPS using a 9600 baud connection (the default for most
  // GPS modules). 
  gps.begin(9600);

  //GPS is pre-configured, but in volatile memory - if the backup batt fails, needs to be reset
  // Configure GPS to onlu output minimum data (location, time, fix).
  //gps.sendCommand(PMTK_SET_NMEA_OUTPUT_RMCONLY);
  //gps.sendCommand("$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29");
  
  Wire.begin();
  Serial.begin(9600);
  lcdSerial.begin(9600); // set up serial port for 9600 baud
 
  enableGPSInterrupt();
  
  pinMode(0, INPUT_PULLUP); //Digital pin 0 (RX)
  pinMode(1, INPUT_PULLUP); //Digital pin 1 (TX)
  attachInterrupt(digitalPinToInterrupt(1), gps_pps_interrupt, RISING);
  attachInterrupt(digitalPinToInterrupt(0), atomic_pps_interrupt, RISING);
 
  // set the initial time here:
  // DS3231 seconds, minutes, hours, day, date, month, year
  // setDS3231time(00,56,13,4,14,9,16);
    // Enable the interrupt to parse GPS data.
  
}

void gps_pps_interrupt(){
  unsigned long  now = micros();
  //check to see if pps_interrupt() was called in the last 250 milliseconds
  if (now - pps_start_ms > 600)
  {
    gps_flag = true;
    //Serial.print("pin 0");
    //Serial.print(now - pps_start_ms);
    //Serial.print(", ");
    pps_start_ms = now;
  }
}

void atomic_pps_interrupt(){
  unsigned long  now = micros();
    if (now - rtc_start_ms > 600)
  {
    atomic_flag = true;
    //Serial.print("pin 1");
    //Serial.println(now - rtc_start_ms);
    rtc_start_ms = now;
  }
}

 
//i2c switch control
void tcaselect(uint8_t i) {
  if (i > 7) return;
 
  Wire.beginTransmission(TCAADDR);
  Wire.write(1 << i);
  Wire.endTransmission();  
}

// Convert normal decimal numbers to binary coded decimal
byte decToBcd(byte val)
{
  return( (val/10*16) + (val%10) );
}
// Convert binary coded decimal to normal decimal numbers
byte bcdToDec(byte val)
{
  return( (val/16*10) + (val%16) );
}

void setDS3231time(byte second, byte minute, byte hour, byte dayOfWeek, byte
dayOfMonth, byte month, byte year)
{
  // sets time and date data to DS3231
  Wire.beginTransmission(DS3231_I2C_ADDRESS);
  Wire.write(0); // set next input to start at the seconds register
  Wire.write(decToBcd(second)); // set seconds
  Wire.write(decToBcd(minute)); // set minutes
  Wire.write(decToBcd(hour)); // set hours
  Wire.write(decToBcd(dayOfWeek)); // set day of week (1=Sunday, 7=Saturday)
  Wire.write(decToBcd(dayOfMonth)); // set date (1 to 31)
  Wire.write(decToBcd(month)); // set month
  Wire.write(decToBcd(year)); // set year (0 to 99)
  Wire.endTransmission();
}
void readDS3231time(byte *second,
byte *minute,
byte *hour,
byte *dayOfWeek,
byte *dayOfMonth,
byte *month,
byte *year)
{
  Wire.beginTransmission(DS3231_I2C_ADDRESS);
  Wire.write(0); // set DS3231 register pointer to 00h
  Wire.endTransmission();
  Wire.requestFrom(DS3231_I2C_ADDRESS, 7);
  // request seven bytes of data from DS3231 starting from register 00h
  *second = bcdToDec(Wire.read() & 0x7f);
  *minute = bcdToDec(Wire.read());
  *hour = bcdToDec(Wire.read() & 0x3f);
  *dayOfWeek = bcdToDec(Wire.read());
  *dayOfMonth = bcdToDec(Wire.read());
  *month = bcdToDec(Wire.read());
  *year = bcdToDec(Wire.read());
}
void displayTime(int clockSource)
{
  byte second, minute, hour, dayOfWeek, dayOfMonth, month, year;

    //Select the i2c channel for the clock source
  tcaselect(clockSource); 
  // retrieve data from the selected DS3231
  readDS3231time(&second, &minute, &hour, &dayOfWeek, &dayOfMonth, &month, &year);

  lcdSerial.write(254); // move cursor to beginning of first line
  lcdSerial.write(128); //move cursor to beginning of first line
  //Print the label and overwrite whatever was on the line
  lcdSerial.print("                    ");
  lcdSerial.write(254); // move cursor to beginning of first line
  lcdSerial.write(128); //move cursor to beginning of first line
  
 // send it to the Serial monitor
  switch(dayOfWeek){
  case 1:
    lcdSerial.print("Sunday     ");
    break;
  case 2:
    lcdSerial.print("Monday     ");
    break;
  case 3:
    lcdSerial.print("Tuesday    ");
    break;
  case 4:
    lcdSerial.print("Wednesday  ");
    break;
  case 5:
    lcdSerial.print("Thursday   ");
    break;
  case 6:
    lcdSerial.print("Friday     ");
    break;
  case 7:
    lcdSerial.print("Saturday   ");
    break;
  }

  lcdSerial.print(month, DEC);
  lcdSerial.print("/");
  lcdSerial.print(dayOfMonth, DEC);
  lcdSerial.print("/20"); //note: should keep track of the century...
  lcdSerial.print(year, DEC);
  
  //Display the proper label for the clock source
  switch(clockSource){
  case REFERENCE_RTC: //0 - Fourth LCD Line
    lcdSerial.write(254); // move cursor to beginning of fourth line
    lcdSerial.write(212); //move cursor to beginning of fourth line
    //Print the label and overwrite whatever was on the line
    lcdSerial.print("                    ");
    lcdSerial.write(254); // move cursor to beginning of fourth line
    lcdSerial.write(212); //move cursor to beginning of fourth line
    lcdSerial.print("Ref :  "); 
    break;
  case GPS_RTC: //1 - Third LCD line
    lcdSerial.write(254); // move cursor to beginning of third line
    lcdSerial.write(148); //move cursor to beginning of third line
    //Print the label and overwrite whatever was on the line
    lcdSerial.print("                    ");
    lcdSerial.write(254); // move cursor to beginning of third line
    lcdSerial.write(148); //move cursor to beginning of third line
    lcdSerial.print("GPS :  ");    
    break; 
  case ATOMIC_RTC: //2 - Second LCD line
    lcdSerial.write(254); // move cursor to beginning of second line
    lcdSerial.write(192); //move cursor to beginning of second line
    //Print the label and overwrite whatever was on the line
    lcdSerial.print("                    ");
    lcdSerial.write(254); // move cursor to beginning of second line
    lcdSerial.write(192); //move cursor to beginning of second line
    lcdSerial.print("Atom:  ");
    break;
  }

  lcdSerial.print(hour, DEC);
  // convert the byte variable to a decimal number when displayed
  lcdSerial.print(":");
  if (minute<10)
  {
    lcdSerial.print("0");
  }
  lcdSerial.print(minute, DEC);
  lcdSerial.print(":");
  if (second<10)
  {
    lcdSerial.print("0");
  }
  lcdSerial.print(second, DEC);
  
  //Print milliseconds for GPS (from the NEMA message, not the RTC module)
  if(clockSource == GPS_RTC){
    lcdSerial.print(":");  
    lcdSerial.print(gps.milliseconds); 
  }
}

SIGNAL(TIMER0_COMPA_vect) {
  // Use a timer interrupt once a millisecond to check for new GPS data.
  // This piggybacks on Arduino's internal clock timer for the millis()
  // function.
  gps.read();
}

void enableGPSInterrupt() {
  // Function to enable the timer interrupt that will parse GPS data.
  // Timer0 is already used for millis() - we'll just interrupt somewhere
  // in the middle and call the "Compare A" function above
  OCR0A = 0xAF;
  TIMSK0 |= _BV(OCIE0A);
}

void loop()
{

  // Check if GPS has new data and parse it.
  if (gps.newNMEAreceived()) {
    gps.parse(gps.lastNMEA());
  }
  
  if(gps_flag){
    //Serial.print(gps.seconds,DEC);
    if(setTime == true){
         // set the initial time here:
        // DS3231 seconds, minutes, hours, day, date, month, year
        //Select the i2c channel for the clock source
        tcaselect(GPS_RTC); 
        setDS3231time(gps.seconds,gps.minute,gps.hour,5,22,9,16);
        tcaselect(ATOMIC_RTC); 
        setDS3231time(gps.seconds,gps.minute,gps.hour,5,22,9,16);
        tcaselect(REFERENCE_RTC); 
        setDS3231time(gps.seconds,gps.minute,gps.hour,5,22,9,16);        
        //setTime=false;
    }
    displayTime(GPS_RTC); // display the real-time clock data on the //Serial Monitor,
    gps_flag = false;
  }

    if(atomic_flag){
    Serial.print(gps.seconds,DEC);
    displayTime(ATOMIC_RTC); // display the real-time clock data on the //Serial Monitor,
    displayTime(REFERENCE_RTC); //Aslo display the Reference RTC time
    atomic_flag = false;
  }
}



